/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Dispatch/dispatcher.h>
#include <Dispatch/iocallback.h>

#include <Graphic/graphicmaster.h>
#include <Graphic/globals.h>
#include <Graphic/glyphviewer.h>
#include <Graphic/noptool.h>

#include <IV-look/kit.h>
#include <IV-look/slider.h>

#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/cursor.h>
#include <InterViews/event.h>
#include <InterViews/handler.h>
#include <InterViews/hit.h>
#include <InterViews/session.h>
#include <InterViews/style.h>
#include <InterViews/target.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>

#include <IV-X11/xcanvas.h>

#include <OS/string.h>

#include <math.h>
#include <stream.h>

//
// GlyphGrabber: a helper class for handling inputs
//

class GlyphGrabber: public Handler
{
	public:
		GlyphGrabber(GlyphViewer *,Allocation *);
		virtual boolean event(Event &);
	protected:
		GlyphViewer *_gviewer;
		Allocation *_a;
};

GlyphGrabber::GlyphGrabber(GlyphViewer *gviewer,Allocation *a)
{
	_gviewer=gviewer;
	_a=a;
}

boolean GlyphGrabber::event(Event &e)
{
	Event ev=e;
	Session *s=Session::instance();
	Graphic *root=_gviewer->root();
	Tool *tool=_gviewer->tool();
	Canvas *c=e.window()->canvas();
	boolean ok;
	
	c->damage_all();
	c->push_clipping();
	//*** setting the GraphicMaster's boundary as clipping rectangle ***
	c->clip_rect(_a->left(),_a->bottom(),_a->right(),_a->top());
	c->rep()->start_repair();
	c->rep()->finish_repair();
	//******************************************************************
	if(e.pointer_button()==Event::middle)
	{
		if(!tool->button_used(GlyphViewer::middle) && _gviewer->tool(GlyphViewer::middle)!=nil)
			tool=_gviewer->tool(GlyphViewer::middle);
	}
	else
		if(e.pointer_button()==Event::right)
		{
			if(!tool->button_used(GlyphViewer::right) && _gviewer->tool(GlyphViewer::right)!=nil)
				tool=_gviewer->tool(GlyphViewer::right);
		}
	if(tool!=nil)
	{
		ok=tool->grasp(ev);
		while(ok)
		{
			s->read(ev);
			ok=tool->manipulating(ev);
		}
		ok=tool->effect(ev);
	}
	c->pop_clipping();
	return ok;
}	

//

const static Coord lower_xy=0.0;
const static Coord upper_xy=100.0;
const static Coord cur_begin_xy=45.0;
const static Coord cur_end_xy=55.0;
const static Coord cur_length_xy=cur_end_xy-cur_begin_xy;
const static Coord width_xy=upper_xy-lower_xy-cur_length_xy;
//const static Coord scaling_xy=upper_xy/cur_length_xy;

//
// GraphicZoomer
//

class GraphicZoomer: public Adjustable
{
	public:
		GraphicZoomer(GlyphViewer *);
		virtual ~GraphicZoomer();

		GlyphViewer *glyphviewer() const;

		virtual void begin_adjustment(DimensionName);
		virtual void commit_adjustment(DimensionName);

		virtual Coord lower(DimensionName) const;
		virtual Coord upper(DimensionName) const;
		virtual Coord cur_lower(DimensionName) const;
		virtual Coord cur_upper(DimensionName) const;
		virtual void scroll_to(DimensionName,Coord);

		virtual void scroll_forward(DimensionName d);
		virtual void scroll_backward(DimensionName d);
		virtual void page_forward(DimensionName d);
		virtual void page_backward(DimensionName d);

		float scale() const {return _scale;}
	private:
		GlyphViewer *_gv;
		Coord _curl,_curr;
		float _scale;
};

GraphicZoomer::GraphicZoomer(GlyphViewer *gv)
{
	_gv=gv;
	Resource::ref(_gv);
	_curl=cur_begin_xy;
	_curr=cur_end_xy;
	_scale=1.0;
	large_scroll(Dimension_X,30.0);
	large_scroll(Dimension_Y,30.0);
	small_scroll(Dimension_X,10.0);
	small_scroll(Dimension_Y,10.0);
}

GraphicZoomer::~GraphicZoomer()
{
	Resource::unref(_gv);
}

GlyphViewer *GraphicZoomer::glyphviewer() const
{
	return _gv;
}

void GraphicZoomer::begin_adjustment(DimensionName d)
{
}

void GraphicZoomer::commit_adjustment(DimensionName d)
{
	float r;
	Coord x,y;
	GraphicMaster *gm=_gv->root();
	Transformer *tf=gm->transformer();
	
	if(tf!=nil)
		tf->transform(0.0,0.0,x,y);
	else
		x=y=0;
	if(d==Dimension_X)
	{
		r=pow(2,3*(2.0*_curl-width_xy)/width_xy)/_scale;
		gm->scale(r,r,x,y);
		_scale*=r;
	}
	_gv->panner()->allocate_thumb(_gv->panner()->allocation());
	gm->update();
	_gv->damage_all();
}

Coord GraphicZoomer::lower(DimensionName) const
{
	return lower_xy;
}

Coord GraphicZoomer::upper(DimensionName) const
{
	return upper_xy;
}

Coord GraphicZoomer::cur_lower(DimensionName d) const
{
	if(d==Dimension_X)
		return _curl;
	return 0;
}

Coord GraphicZoomer::cur_upper(DimensionName d) const
{
	if(d==Dimension_X)
		return _curr;
	return 0;
}

void GraphicZoomer::scroll_to(DimensionName d,Coord pos)
{
	constrain(d,pos);
	if(d==Dimension_X)
	{
		if(pos!=_curl)
		{
			_curl=pos;
			_curr=pos+cur_length_xy;
		}
		notify(d);
	}
}

void GraphicZoomer::scroll_forward(DimensionName d)
{
	Adjustable::scroll_forward(d);
	commit_adjustment(d);
}

void GraphicZoomer::scroll_backward(DimensionName d)
{
	Adjustable::scroll_backward(d);
	commit_adjustment(d);
}

void GraphicZoomer::page_forward(DimensionName d)
{
	Adjustable::page_forward(d);
	commit_adjustment(d);
}

void GraphicZoomer::page_backward(DimensionName d)
{
	Adjustable::page_backward(d);
	commit_adjustment(d);
}

//
// GlyphViewer
//

declareIOCallback(GlyphViewer)
implementIOCallback(GlyphViewer)

GlyphViewer::GlyphViewer(Coord w,Coord h,Coord pw,Coord ph):
	InputHandler(new Target(_root=new GraphicMaster(nil),TargetAlwaysHit),Session::instance()->style())
{
	Color *mg=new Color(1.0,1.0,1.0,1.0);

	_root->glyphviewer(this);
	_root->background(mg->brightness(0.0));
	delete mg;
	_width=w;
	_height=h;
	_grabber=new GlyphGrabber(this,&_a);
	_grabber->ref();
	_tool[GlyphViewer::left]=new NOPTool(_root);
	_cursor=nil;
	Resource::ref(_tool[GlyphViewer::left]);
	_tool[GlyphViewer::middle]=_tool[GlyphViewer::right]=nil;
	_zoomer=new GraphicZoomer(this);

	_canvas=nil;
	_pagew=pw/2;
	_pageh=ph/2;
	_curl=-w/2.0;
	_curr=-_curl;
	_curb=-h/2.0;
	_curt=-_curt;
	_panner=nil;
	initshape();
	initgraphic();
	inittimer();
}

GlyphViewer::~GlyphViewer()
{
	delete _timer;
	_grabber->unref();
	delete _zoomer;
	Resource::unref(_panner);
}

void GlyphViewer::press(const Event &e)
{
	Window *w=canvas()->window();
	_cursor=w->cursor();
	Dispatcher::instance().startTimer(0,_delay/40,_timer);
}

void GlyphViewer::drag(const Event &e)
{
}

void GlyphViewer::release(const Event&)
{
	Window *w=canvas()->window();
	w->cursor(_cursor);
	Dispatcher::instance().stopTimer(_timer);
}

void GlyphViewer::request(Requisition &req) const
{
	Requirement &rx=req.x_requirement();
	rx.natural(_width);
	rx.stretch(fil);
	rx.shrink(_width);
	rx.alignment(0.0);
	
	Requirement &ry=req.y_requirement();
	ry.natural(_height);
	ry.stretch(fil);
	ry.shrink(0.0);
	ry.alignment(0.0);
}

void GlyphViewer::allocate(Canvas *c,const Allocation &a,Extension &ext)
{
	Coord w,h;

	_canvas=c;
	_a=a;
	InputHandler::allocate(c,a,ext);
	ext.merge(c,a);
	scroll_to(Dimension_X,_curl);
	scroll_to(Dimension_Y,_curb);
	w=_a.right()-_a.left();
	h=_a.top()-_a.bottom();
	small_scroll(Dimension_X,w/4);
	small_scroll(Dimension_Y,h/4);
	large_scroll(Dimension_X,w);
	large_scroll(Dimension_Y,h);
}

void GlyphViewer::draw(Canvas *c,const Allocation &a) const
{
	_root->draw(c,a);
}

void GlyphViewer::pick(Canvas *c,const Allocation &a,int depth,Hit &h)
{
	const Event *e=h.event();

	_canvas=c;
	if(e->type()==Event::down)
	{
		h.begin(depth,this,0,_grabber);
		MonoGlyph::pick(c,a,depth,h);
		h.end();
	}
	else
		InputHandler::pick(c,a,depth,h);
}

GraphicMaster *GlyphViewer::root() const
{
	return _root;
}

void GlyphViewer::root(GraphicMaster *root)
{
	if(_root!=root)
	{
		Extension ext;
		Allocation a;

		root->transformer(nil);
		root->background(_root->background());
		_root->allocate(nil,a,ext);
		_root=root;
		body(new Target(_root,TargetAlwaysHit));
		initgraphic();
		Canvas *c=canvas();
		if(c!=nil)
		{
			Requisition req;
			_root->request(req);
			_root->allocate(c,_a,ext);
		}
		_root->glyphviewer(this);
	}
}

Tool *GlyphViewer::tool(unsigned i) const
{
	if(i>GlyphViewer::right)
		i=GlyphViewer::left;
	return _tool[i];
}

void GlyphViewer::tool(Tool *tool,unsigned i)
{
	if(i>GlyphViewer::right)
		i=GlyphViewer::left;
	Resource::ref(tool);
	Resource::unref(_tool[i]);
	_tool[i]=tool;
}

void GlyphViewer::damage(Coord l,Coord b,Coord r,Coord t)
{
	Canvas *c=canvas();

	if(c!=nil)
		c->damage(l,b,r,t);
}

void GlyphViewer::damage_all()
{
	Canvas *c=canvas();

	if(c!=nil)
		c->damage_all();
}

void GlyphViewer::repair()
{
	Canvas *c=canvas();

	if(c!=nil)
		c->window()->repair();
}

void GlyphViewer::clip_all() const
{
        _canvas->clip_rect(_a.left(),_a.bottom(),_a.right(),_a.top());
}

void GlyphViewer::begin_adjustment(DimensionName d)
{
}

void GlyphViewer::commit_adjustment(DimensionName d)
{
	Coord x,y,cx,cy;
	Coord a00,a01,a10,a11,a20,a21;
	GraphicMaster *gm=root();
	Transformer *tf=gm->transformer();
	
	if(tf==nil)
	{
		tf=new Transformer;
		gm->transformer(tf);
		a00=a11=1.0;
		a01=a10=a20=a21=0.0;
		tf->unref();
	}
	else
		tf->matrix(a00,a01,a10,a11,a20,a21);
	gm->centerpoint(cx,cy);
	if(d==Dimension_X || d==Dimension_Y)
	{
		x=-(_curl+_curr)/2;
		y=-(_curb+_curt)/2;
		gm->translate(cx+x-a20,cy+y-a21);
		gm->update();
	}
	damage_all();
}

Coord GlyphViewer::lower(DimensionName d) const
{
	if(d==Dimension_X)
		return -_pagew*_zoomer->scale();
	if(d==Dimension_Y)
		return -_pageh*_zoomer->scale();
}

Coord GlyphViewer::upper(DimensionName d) const
{
	if(d==Dimension_X)
		return _pagew*_zoomer->scale();
	if(d==Dimension_Y)
		return _pageh*_zoomer->scale();
	return 0;
}

Coord GlyphViewer::length(DimensionName d) const
{
	if(d==Dimension_X)
		return 2*_pagew*_zoomer->scale();
	if(d==Dimension_Y)
		return 2*_pageh*_zoomer->scale();
	return 0;
}

Coord GlyphViewer::cur_lower(DimensionName d) const
{
	if(d==Dimension_X)
		return _curl;
	if(d==Dimension_Y)
		return _curb;
	return 0;
}

Coord GlyphViewer::cur_upper(DimensionName d) const
{
	if(d==Dimension_X)
		return _curr;
	if(d==Dimension_Y)
		return _curt;
	return 0;
}

Coord GlyphViewer::cur_length(DimensionName d) const
{
	if(d==Dimension_X)
		return _a.right()-_a.left();
	if(d==Dimension_Y)
		return _a.top()-_a.bottom();
	return 0;
}

void GlyphViewer::scroll_to(DimensionName d,Coord pos)
{
	constrain(d,pos);
	if(d==Dimension_X)
	{
		_curl=pos;
		_curr=pos+cur_length(Dimension_X);
		notify(d);
	}
	if(d==Dimension_Y)
	{
		_curb=pos;
		_curt=pos+cur_length(Dimension_Y);
		notify(d);
	}
}

void GlyphViewer::scroll_forward(DimensionName d)
{
	Adjustable::scroll_forward(d);
	commit_adjustment(d);
}

void GlyphViewer::scroll_backward(DimensionName d)
{
	Adjustable::scroll_backward(d);
	commit_adjustment(d);
}

void GlyphViewer::page_forward(DimensionName d)
{
	Adjustable::page_forward(d);
	commit_adjustment(d);
}

void GlyphViewer::page_backward(DimensionName d)
{
	Adjustable::page_backward(d);
	commit_adjustment(d);
}

Slider *GlyphViewer::panner() const
{
	return _panner;
}

void GlyphViewer::panner(Slider *panner)
{
	if(_panner!=nil)
	{
		zoomer()->detach(Dimension_X,_panner);
	}
	Resource::ref(panner);
	Resource::unref(_panner);
	_panner=panner;
	if(_panner!=nil)
	{
		zoomer()->attach(Dimension_X,_panner);
	}
}

// protected operations

void GlyphViewer::initshape()
{
	if(_width<-0.5 && _height<-0.5)
	{
		Coord l,b,r,t;

		_root->getbounds(l,b,r,t);
		_width=r-l;
		_height=t-b;
	}
	Allotment &ax=_a.x_allotment();
	Allotment &ay=_a.y_allotment();
	ax.span(_width);
	ay.span(_height);
}

void GlyphViewer::initgraphic()
{
	Coord l,b,r,t;
	Canvas *c=canvas();

	//_root->getbounds(l,b,r,t);
	//_root->translate(-l-(r-l)/2.0,-b-(t-b)/2.0);
	if(c!=nil)
		c->damage_all();
}

void GlyphViewer::inittimer()
{
	float seconds=2.0;
	_delay=long(seconds  *1000000);
	_timer=new IOCallback(GlyphViewer)(this,&GlyphViewer::tick);
}

void GlyphViewer::do_scroll()
{
}

// private operations

void GlyphViewer::tick(long,long)
{
	Dispatcher::instance().startTimer(0,_delay/40,_timer);
}

