/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/globals.h>
#include <Graphic/handles.h>

#include <InterViews/canvas.h>
#include <InterViews/color.h>

#include <string.h>

Handles::Handles(const Coord *x,const Coord *y,int n,Canvas *c)
{
	_canvas=c;
	_x=new Coord[_nhandles=n];
	_y=new Coord[n];
	_drawn=false;
	memcpy(_x,x,sizeof(Coord)*n);
	memcpy(_y,y,sizeof(Coord)*n);
}

Handles::~Handles()
{
	delete _x;
	delete _y;
}

static const Coord epsilon=3;

void Handles::draw(Canvas *c)
{
	if(c!=nil)
		_canvas=c;
	if(!_drawn)
	{
		Color *col=new Color(0,0,0,1,Color::Xor);
		int i;
	
		Resource::ref(col);
		for(i=0;i<_nhandles;i++)
		{
			_canvas->fill_rect(_x[i]-epsilon,_y[i]-epsilon,_x[i]+epsilon,_y[i]+epsilon,col);
		}
		_drawn=true;
		Resource::unref(col);
	}
}

void Handles::redraw()
{
	_drawn=false;
	draw();
}

void Handles::erase()
{
	if(_drawn)
	{
		_drawn=false;
		draw();
		_drawn=false;
	}
}

void Handles::getbounds(Coord &l,Coord &b,Coord &r,Coord &t)
{
	if(_nhandles>0)
	{
		int i;

		l=r=_x[0];
		b=t=_y[0];
		for(i=1;i<_nhandles;i++)
		{
			if(_x[i]<l)
				l=_x[i];
			else
				if(_x[i]>r)
					r=_x[i];
			if(_y[i]<b)
				b=_y[i];
			else
				if(_y[i]>t)
					t=_y[i];
		}
		l-=epsilon;
		b-=epsilon;
		r+=epsilon;
		t+=epsilon;
	}
}

Coord Handles::tol()
{
	return epsilon;
}

