/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/splines.h>

#include <InterViews/brush.h>
#include <InterViews/color.h>
#include <InterViews/transformer.h>

//
//
//

Open_BSpline::Open_BSpline(const Brush *brush,const Color *stroke,const Color *fill,Coord *x,Coord *y,int n,Transformer *t):
	Graphic(brush,stroke,fill,nil,false,true,(n+2)*3+1,t)
{
	int i;

	Bspline_move_to(x[0],y[0],x[0],y[0],x[0],y[0]);
	Bspline_curve_to(x[0],y[0],x[0],y[0],x[1],y[1]);
	for(i=1;i<n-1;++i)
		Bspline_curve_to(x[i],y[i],x[i-1],y[i-1],x[i+1],y[i+1]);
	Bspline_curve_to(x[n-1],y[n-1],x[n-2],y[n-2],x[n-1],y[n-1]);
	Bspline_curve_to(x[n-1],y[n-1],x[n-1],y[n-1],x[n-1],y[n-1]);
}

Open_BSpline::Open_BSpline(Open_BSpline *gr): Graphic(gr)
{
}

Open_BSpline::~Open_BSpline()
{
}

Glyph *Open_BSpline::clone() const
{
	return new Open_BSpline((Open_BSpline*) this);
}

ClassId Open_BSpline::get_classid() const
{
	return OBSPLINE_GRAPHIC;
}

boolean Open_BSpline::is_a(ClassId id) const
{
	return id==OBSPLINE_GRAPHIC || Graphic::is_a(id);
}

//
//
//

Closed_BSpline::Closed_BSpline(const Brush *brush,const Color *stroke,const Color *fill,Coord *x,Coord *y,int n,Transformer *t):
	Graphic(brush,stroke,fill,nil,true,true,n*3+1,t)
{
	int i;

	Bspline_move_to(x[0],y[0],x[n-1],y[n-1],x[1],y[1]);
	for(i=1;i<n-1;++i)
		Bspline_curve_to(x[i],y[i],x[i-1],y[i-1],x[i+1],y[i+1]);
	Bspline_curve_to(x[n-1],y[n-1],x[n-2],y[n-2],x[0],y[0]);
	Bspline_curve_to(x[0],y[0],x[n-1],y[n-1],x[1],y[1]);
}

Closed_BSpline::Closed_BSpline(Closed_BSpline *gr): Graphic(gr)
{
}

Closed_BSpline::~Closed_BSpline()
{
}

Glyph *Closed_BSpline::clone() const
{
	return new Closed_BSpline((Closed_BSpline*) this);
}

ClassId Closed_BSpline::get_classid() const
{
	return CBSPLINE_GRAPHIC;
}

boolean Closed_BSpline::is_a(ClassId id) const
{
	return id==CBSPLINE_GRAPHIC || Graphic::is_a(id);
}

